#include <stdio.h>
#include <stdlib.h>

#define EDGE 1
#define RELEASE char*

char *cVector(long nfirst, long nlast)
{
   char *vtr;
   vtr = (char *)malloc( (size_t) ((nlast-nfirst+1+EDGE)*sizeof(char)) );
   if(!vtr) printf( "Allocation failed!\n" );
   return vtr-nfirst+EDGE;
}

int *iVector(long nfirst, long nlast)
{
   int *vtr;
   vtr = (int *)malloc( (size_t) ((nlast-nfirst+1+EDGE)*sizeof(int)) );
   if(!vtr) printf( "Allocation failed!\n" );
   return vtr-nfirst+EDGE;
}

double *dVector(long nfirst, long nlast)
{
   double *vtr;
   vtr = (double *)malloc( (size_t) ((nlast-nfirst+1+EDGE)*sizeof(double)) );
   if(!vtr) printf( "Allocation failed!\n" );
   return vtr-nfirst+EDGE;
}

float *fVector(long nfirst, long nlast)
{
   float *vtr;
   vtr = (float *)malloc( (size_t) ((nlast-nfirst+1+EDGE)*sizeof(float)) );
   if(!vtr) printf( "Allocation failed!\n" );
   return vtr-nfirst+EDGE;
}

double **dMatrix(long nrowfirst, long nrowlast, long ncolumnfirst, long ncolumnlast)
{
   long   i, nrow=nrowlast-nrowfirst+1, ncol=ncolumnlast-ncolumnfirst+1;
   double **mtr;

   mtr = (double **) malloc( (size_t)((nrow+EDGE)*sizeof(double*)) );
   if(!mtr) printf( "Allocation failed!\n" );
   mtr += EDGE;
   mtr -= nrowfirst;

   mtr[nrowfirst] = (double *) malloc( (size_t)( (nrow*ncol+EDGE)*sizeof(double)));
   if(!mtr[nrowfirst]) printf( "Allocation failed!\n" );
   mtr[nrowfirst] += EDGE;
   mtr[nrowfirst] -= ncolumnfirst;

   for(i=nrowfirst+1;i<=nrowlast;i++) mtr[i]=mtr[i-1]+ncol;

   return mtr;
}

float **fMatrix(long nrowfirst, long nrowlast, long ncolumnfirst, long ncolumnlast)
{
   long   i, nrow=nrowlast-nrowfirst+1, ncol=ncolumnlast-ncolumnfirst+1;
   float **mtr;

   mtr = (float **) malloc( (size_t)((nrow+EDGE)*sizeof(float*)) );
   if(!mtr) printf( "Allocation failed!\n" );
   mtr += EDGE;
   mtr -= nrowfirst;

   mtr[nrowfirst] = (float *) malloc( (size_t)( (nrow*ncol+EDGE)*sizeof(float)));
   if(!mtr[nrowfirst]) printf( "Allocation failed!\n" );
   mtr[nrowfirst] += EDGE;
   mtr[nrowfirst] -= ncolumnfirst;

   for(i=nrowfirst+1;i<=nrowlast;i++) mtr[i]=mtr[i-1]+ncol;

   return mtr;
}

int **iMatrix(long nrowfirst, long nrowlast, long ncolumnfirst, long ncolumnlast)
{
   long   i, nrow=nrowlast-nrowfirst+1, ncol=ncolumnlast-ncolumnfirst+1;
   int **mtr;

   mtr = (int **) malloc( (size_t)((nrow+EDGE)*sizeof(int*)) );
   if(!mtr) printf( "Allocation failed!\n" );
   mtr += EDGE;
   mtr -= nrowfirst;

   mtr[nrowfirst] = (int *) malloc( (size_t)( (nrow*ncol+EDGE)*sizeof(int)));
   if(!mtr[nrowfirst]) printf( "Allocation failed!\n" );
   mtr[nrowfirst] += EDGE;
   mtr[nrowfirst] -= ncolumnfirst;

   for(i=nrowfirst+1;i<=nrowlast;i++) mtr[i]=mtr[i-1]+ncol;

   return mtr;
}

char **cMatrix(long nrowfirst, long nrowlast, long ncolumnfirst, long ncolumnlast)
{
   long   i, nrow=nrowlast-nrowfirst+1, ncol=ncolumnlast-ncolumnfirst+1;
   char **mtr;

   mtr = (char **) malloc( (size_t)((nrow+EDGE)*sizeof(char*)) );
   if(!mtr) printf( "Allocation failed!\n" );
   mtr += EDGE;
   mtr -= nrowfirst;

   mtr[nrowfirst] = (char *) malloc( (size_t)( (nrow*ncol+EDGE)*sizeof(char)));
   if(!mtr[nrowfirst]) printf( "Allocation failed!\n" );
   mtr[nrowfirst] += EDGE;
   mtr[nrowfirst] -= ncolumnfirst;

   for(i=nrowfirst+1;i<=nrowlast;i++) mtr[i]=mtr[i-1]+ncol;

   return mtr;
}

void free_iVector(int *vtr, long nfirst, long nlast)
{
	free( (RELEASE) (vtr+nfirst-EDGE) );
}

void free_cVector(char *vtr, long nfirst, long nlast)
{
	free( (RELEASE) (vtr+nfirst-EDGE) );
}

void free_dVector(double *vtr, long nfirst, long nlast)
{
	free( (RELEASE) (vtr+nfirst-EDGE) );
}

void free_fVector(float *vtr, long nfirst, long nlast)
{
	free( (RELEASE) (vtr+nfirst-EDGE) );
}

void free_dMatrix(double **mtr, long nrowfirst, long nrowlast, long ncolumnfirst, long ncolumnlast)
{
    free( (RELEASE) (mtr[nrowfirst]+ncolumnfirst-EDGE) );
    free( (RELEASE) (mtr+nrowfirst-EDGE) );
}

void free_fMatrix(float **mtr, long nrowfirst, long nrowlast, long ncolumnfirst, long ncolumnlast)
{
    free( (RELEASE) (mtr[nrowfirst]+ncolumnfirst-EDGE) );
    free( (RELEASE) (mtr+nrowfirst-EDGE) );
}

void free_iMatrix(int **mtr, long nrowfirst, long nrowlast, long ncolumnfirst, long ncolumnlast)
{
    free( (RELEASE) (mtr[nrowfirst]+ncolumnfirst-EDGE) );
    free( (RELEASE) (mtr+nrowfirst-EDGE) );
}

void free_cMatrix(char **mtr, long nrowfirst, long nrowlast, long ncolumnfirst, long ncolumnlast)
{
    free( (RELEASE) (mtr[nrowfirst]+ncolumnfirst-EDGE) );
    free( (RELEASE) (mtr+nrowfirst-EDGE) );
}