#include <math.h>
#include <stdio.h>


#include "umath.h"



typedef double rotate_t[3][3];

typedef double xlate_t[3];

typedef struct Coord { double x, y, z; } CoordType;



int     *iVector(long, long);

int    **iMatrix(long, long, long, long);

double  *dVector(long, long);

char    *cVector(long, long);

double **dMatrix(long, long, long, long);

void free_iVector(int     *, long, long);

void free_iMatrix(int    **, long, long, long, long);

void free_dVector(double  *, long, long);

void free_dMatrix(double **, long, long, long, long);

void free_cVector(char    *, long, long);



/*=====================================================================*/

static int solve(double e[3],

                   double e0, double det, double spur, double cof)

{

double d, h, g, cth, sth, sqrth;

#define epsilon 1e-7

#define sqrt3 1.7320508075688772

  

  det *= det;

  d = SQR(spur);

  h = d - cof;

  g = spur * (cof*1.5 - d) - det*0.5;

  if (h > d*epsilon) {

    sqrth = SQRTABS(h);

    d = -g/(h*sqrth);

    if (d > 1-epsilon) {    /* Two identical roots */

      e[0] = spur + 2*sqrth;

      e[2] = e[1] = spur - sqrth;

      return 2;

      }

    else if (d < -1+epsilon) {    /* Two identical roots */ 

      e[1] = e[0] = spur+sqrth;

      e[2] = MAX(0, spur - 2*sqrth);

      return 3;

      }

    else {    /* Three distinct roots */

      d = acos(d)/3.0;

      cth = sqrth*cos(d); sth = sqrth*sqrt3*sin(d);

      e[0] = spur + 2*cth;

      e[1] = spur - cth + sth;

      e[2] = MAX(0, spur - cth - sth);

      return 1;

      }

    }

  else {    /* Three identical roots */ 

    e[0] = e[1] = e[2] = spur;

    return 4;

    }



 } /* solve */





static int matrix(rotate_t u, xlate_t t,

                   int sw, double e[3], double rr[6], double r[3][3],

                   double sx[3], double sy[3], int npts)

{

  double ss[6], d, h, p, a[3][3], b[3][3];

  int i, j, l, m, m1, m2, m3;



  if (sw == 4) {

    /* Case of three identical roots */

    for (i = 0; i < 3; i++)

      for (j = 0; j < 3; j++)

        if (i == j) a[i][j] = 1.0; else a[i][j] = 0.0;

    }

  else {

    if (sw == 1) {

      /* Case of three distinct roots */

      for (l=0; l < 2; l++) {

        d = e[l];

        ss[0] = (d-rr[2]) * (d-rr[5]) - SQR(rr[4]);

        ss[1] = (d-rr[5]) * rr[1] + rr[3]*rr[4];

        ss[2] = (d-rr[0]) * (d-rr[5]) - SQR(rr[3]);

        ss[3] = (d-rr[2]) * rr[3] + rr[1]*rr[4];

        ss[4] = (d-rr[0]) * rr[4] + rr[1]*rr[3];

        ss[5] = (d-rr[0]) * (d-rr[2]) - SQR(rr[1]);

        if (FABS(ss[0]) >= FABS(ss[2])) {

          if (FABS(ss[0]) >= FABS(ss[5])) {

            a[0][l] = ss[0]; a[1][l] = ss[1]; a[2][l] = ss[3];

            }

          else {

            a[0][l] = ss[3]; a[1][l] = ss[4]; a[2][l] = ss[5];

            }

          }

        else if (FABS(ss[2]) >= FABS(ss[5])) {

          a[0][l] = ss[1]; a[1][l] = ss[2]; a[2][l] = ss[4];

          }

        else {

          a[0][l] = ss[3]; a[1][l] = ss[4]; a[2][l] = ss[5];

          }

        d = sqrt(SQR(a[0][l]) + SQR(a[1][l]) + SQR(a[2][l]));

        a[0][l] /= d; a[1][l] /= d; a[2][l] /= d;

        }

      m1 = 2; m2 = 0; m3 = 1;

      }

    else {

      /* Cases of two distinct roots */

      if (sw == 2) {

        m = 0; m1 = 2; m2 = 0; m3 = 1;

        }

      else {

        m = 2; m1 = 0; m2 = 1; m3 = 2;

        }

      h = e[2];

      a[0][1] = 1.0; a[1][1] = 1.0; a[2][1] = 1.0;

      if (FABS(rr[0]-h) > FABS(rr[2]-h)) {

        if (FABS(rr[0]-h) > FABS(rr[5]-h)) {

          a[0][m] = rr[0]-h; a[1][m] = rr[1]; a[2][m] = rr[3];

          p = -(rr[0] + rr[1] + rr[3] - h);

          a[0][1] = p/a[0][m];

          }

        else {

          a[0][m] = rr[3]; a[1][m] = rr[4]; a[2][m] = rr[5]-h;

          p = -(rr[3] + rr[4] + rr[5] - h);

          a[2][1] = p/a[2][m];

          }

        }

      else {

        if (FABS(rr[2]-h) > FABS(rr[5]-h)) {

          a[0][m] = rr[1]; a[1][m] = rr[2]-h; a[2][m] = rr[4];

          p = -(rr[1] + rr[2] + rr[4] - h);

          a[1][1] = p/a[1][m];

          }

        else {

          a[0][m] = rr[3]; a[1][m] = rr[4]; a[2][m] = rr[5]-h;

          p = -(rr[3] + rr[4] + rr[5] - h);

          a[2][1] = p/a[2][m];

          }

        }

      d = sqrt(SQR(a[0][m]) + SQR(a[1][m]) + SQR(a[2][m]));

      p = sqrt(SQR(a[0][1]) + SQR(a[1][1]) + SQR(a[2][1]));

      for (i = 0; i < 3; i++) {

        a[i][1] /= p;

        a[i][m] /= d;

        }

      }

    /* Common for either two or three distinct roots */

    a[0][m1] = a[1][m2]*a[2][m3] - a[1][m3]*a[2][m2];

    a[1][m1] = a[2][m2]*a[0][m3] - a[2][m3]*a[0][m2];

    a[2][m1] = a[0][m2]*a[1][m3] - a[0][m3]*a[1][m2];

    }

  

  for (l = 0; l < 2; l++) {

    d = 0;

    for (i = 0; i < 3; i++) {

      b[i][l] = r[0][i]*a[0][l] + r[1][i]*a[1][l] + r[2][i]*a[2][l];

      d += SQR(b[i][l]);

      }

    d = sqrt(d);

    if (d == 0) printf("degenerate rotation matrix in matrix()\n"); 

    for (i = 0; i < 3; i++) b[i][l] /= d;

    }

  b[0][2] = b[1][0]*b[2][1] - b[1][1]*b[2][0];

  b[1][2] = b[2][0]*b[0][1] - b[2][1]*b[0][0];

  b[2][2] = b[0][0]*b[1][1] - b[0][1]*b[1][0];

  

  /* Calculate rotation matrix */

  for (i = 0; i < 3; i++)

    for (j = 0; j < 3; j++)

      u[i][j] = b[i][0]*a[j][0] + b[i][1]*a[j][1] + b[i][2]*a[j][2];

  /* Calculate translation vector */

  for (i = 0; i < 3; i++)

    t[i] = (sy[i] - u[i][0]*sx[0] - u[i][1]*sx[1] - u[i][2]*sx[2])/npts;



  return 7;

}


double bestfit(rotate_t u, xlate_t t, CoordType *a, CoordType *b, const int npts)
{
  double sx[3], sy[3], sxy[3][3], sx2, sy2, temp; 
  double r[3][3], rr[6];
  double d, spur, det, cof, e0, e[3], inpts;
  int i, j, m, sw;


  double vsumby3(double [], int);

  double vsumsqr(double [], int);

  double vdotby3(double [], double [], int n);



  temp = vsumby3(&a[0].x, npts);

  temp = vsumby3(&a[0].y, npts);

  temp = vsumby3(&a[0].z, npts);

  temp = vsumby3(&b[0].x, npts);

  temp = vsumby3(&b[0].y, npts);

  temp = vsumby3(&b[0].z, npts);



  temp = vsumsqr(&a[0].x, 3*npts);

  temp = vsumsqr(&b[0].x, 3*npts);



  temp = vdotby3(&a[0].x, &b[0].x, npts);

  temp = vdotby3(&a[0].x, &b[0].y, npts);

  temp = vdotby3(&a[0].x, &b[0].z, npts);

  temp = vdotby3(&a[0].y, &b[0].x, npts);

  temp = vdotby3(&a[0].y, &b[0].y, npts);

  temp = vdotby3(&a[0].y, &b[0].z, npts);

  temp = vdotby3(&a[0].z, &b[0].x, npts);

  temp = vdotby3(&a[0].z, &b[0].y, npts);

  temp = vdotby3(&a[0].z, &b[0].z, npts);


  sx[0] = vsumby3(&a[0].x, npts);
  sx[1] = vsumby3(&a[0].y, npts);
  sx[2] = vsumby3(&a[0].z, npts);
  sy[0] = vsumby3(&b[0].x, npts);
  sy[1] = vsumby3(&b[0].y, npts);
  sy[2] = vsumby3(&b[0].z, npts);
  sx2   = vsumsqr(&a[0].x, 3*npts);
  sy2   = vsumsqr(&b[0].x, 3*npts);
  sxy[0][0] = vdotby3(&a[0].x, &b[0].x, npts);
  sxy[0][1] = vdotby3(&a[0].x, &b[0].y, npts);
  sxy[0][2] = vdotby3(&a[0].x, &b[0].z, npts);
  sxy[1][0] = vdotby3(&a[0].y, &b[0].x, npts);
  sxy[1][1] = vdotby3(&a[0].y, &b[0].y, npts);
  sxy[1][2] = vdotby3(&a[0].y, &b[0].z, npts);
  sxy[2][0] = vdotby3(&a[0].z, &b[0].x, npts);
  sxy[2][1] = vdotby3(&a[0].z, &b[0].y, npts);
  sxy[2][2] = vdotby3(&a[0].z, &b[0].z, npts);

  inpts = 1.0/npts;
  e0 = (sx2 - (SQR(sx[0])+SQR(sx[1])+SQR(sx[2]))*inpts +
        sy2 - (SQR(sy[0])+SQR(sy[1])+SQR(sy[2]))*inpts)*inpts;

  for (i = 0; i < 3; i++) {
    for (j = 0; j < 3; j++) {
      r[i][j] = (sxy[i][j] - sx[i]*sy[j]*inpts)*inpts;
      }
    }
    
  for (m = j = 0; j < 3; j++) {
    for (i = 0; i <= j; i++,m++) {
      rr[m] = r[i][0]*r[j][0] + r[i][1]*r[j][1] + r[i][2]*r[j][2];
      }
    }

  det = r[0][0] * (r[1][1]*r[2][2] - r[2][1]*r[1][2])
      - r[1][0] * (r[0][1]*r[2][2] - r[2][1]*r[0][2])
      + r[2][0] * (r[0][1]*r[1][2] - r[1][1]*r[0][2]);

  spur = (rr[0]+rr[2]+rr[5])/3.0;
  cof = (rr[2]*rr[5] - SQR(rr[4]) + rr[0]*rr[5] -
         SQR(rr[3]) + rr[0]*rr[2] - SQR(rr[1]))/3.0;

  sw = solve(e, e0, det, spur, cof);
  i = matrix(u, t, sw, e, rr, r, sx, sy, npts);
  d = SQRTABS(e[2]);
  if (det < 0) d = -d;
  d += SQRTABS(e[1]) + SQRTABS(e[0]);
  d = e0-2*d;
  return SQRTABS(d);

  } /* bestfit */

double rotrans(double **x_str, int size_str, rotate_t u, xlate_t t) {

  int i;

  double xx, yy, zz;

  for(i=0; i<size_str; i++) {

    xx = x_str[1][i]; yy = x_str[2][i]; zz = x_str[3][i]; 

    x_str[1][i] = u[0][0]*xx + u[0][1]*yy + u[0][2]*zz;

    x_str[2][i] = u[1][0]*xx + u[1][1]*yy + u[1][2]*zz;

    x_str[3][i] = u[2][0]*xx + u[2][1]*yy + u[2][2]*zz;

    x_str[1][i] +=  t[0];

    x_str[2][i] +=  t[1];

    x_str[3][i] +=  t[2];

  } /* i */

  return 7;

} /* rotrans */



int check_eq(int *iina, int *jinb, int *iina_sav, int *jinb_sav, int nentry) {

  int i, ret_sum;

  ret_sum = 0;

  for(i=0; i<=nentry; i++) {

	if( (iina[i] != iina_sav[i]) || (jinb[i] != jinb_sav[i]) ) ret_sum++;

    if(ret_sum) return ret_sum;

  }

  return ret_sum;

} /* check_eq */



double distrms(CoordType *a, CoordType *b, int npts)

  {

  int i, j;

  double ad, bd, sum;



  sum = 0.0;

  for (i = 0; i < npts; i++) {

    for (j = 0; j < i; j++) {

      ad = sqrt(SQR(a[i].x-a[j].x) + SQR(a[i].y-a[j].y) +

                 SQR(a[i].z-a[j].z));

      bd = sqrt(SQR(b[i].x-b[j].x) + SQR(b[i].y-b[j].y) +

                 SQR(b[i].z-b[j].z));

      sum += SQR(ad-bd);

      }

    }

  return sqrt(2*sum/(npts*(npts-1)));

} /* distrms */



char convert321(char *s) {

 if(s[0] == 'G' && s[1] == 'L' && s[2] == 'Y') return 'G';   

 if(s[0] == 'A' && s[1] == 'L' && s[2] == 'A') return 'A';    

 if(s[0] == 'V' && s[1] == 'A' && s[2] == 'L') return 'V';   

 if(s[0] == 'L' && s[1] == 'E' && s[2] == 'U') return 'L';    

 if(s[0] == 'I' && s[1] == 'L' && s[2] == 'E') return 'I';   

 if(s[0] == 'S' && s[1] == 'E' && s[2] == 'R') return 'S';    

 if(s[0] == 'T' && s[1] == 'H' && s[2] == 'R') return 'T';   

 if(s[0] == 'C' && s[1] == 'Y' && s[2] == 'S') return 'C';    

 if(s[0] == 'M' && s[1] == 'E' && s[2] == 'T') return 'M';   

 if(s[0] == 'P' && s[1] == 'R' && s[2] == 'O') return 'P';    

 if(s[0] == 'A' && s[1] == 'S' && s[2] == 'P') return 'D';   

 if(s[0] == 'A' && s[1] == 'S' && s[2] == 'N') return 'N';    

 if(s[0] == 'G' && s[1] == 'L' && s[2] == 'U') return 'E';   

 if(s[0] == 'G' && s[1] == 'L' && s[2] == 'N') return 'Q';    

 if(s[0] == 'L' && s[1] == 'Y' && s[2] == 'S') return 'K';   

 if(s[0] == 'A' && s[1] == 'R' && s[2] == 'G') return 'R';    

 if(s[0] == 'H' && s[1] == 'I' && s[2] == 'S') return 'H';   

 if(s[0] == 'P' && s[1] == 'H' && s[2] == 'E') return 'F';    

 if(s[0] == 'T' && s[1] == 'Y' && s[2] == 'R') return 'Y';   

 if(s[0] == 'T' && s[1] == 'R' && s[2] == 'P') return 'W';	 

 printf("321 convertion did not succeed! \n");

 return 'X';

} /* convert321 */



double blossom62(char a, char b) {

if( a == 'A' && b == 'A' ) return    8 ;

if( a == 'B' && b == 'A' ) return   -4 ;

if( a == 'B' && b == 'B' ) return    8 ;

if( a == 'C' && b == 'A' ) return    0 ;

if( a == 'C' && b == 'B' ) return   -6 ;

if( a == 'C' && b == 'C' ) return   18 ;

if( a == 'D' && b == 'A' ) return   -4 ;

if( a == 'D' && b == 'B' ) return    8 ;

if( a == 'D' && b == 'C' ) return   -6 ;

if( a == 'D' && b == 'D' ) return   12 ;

if( a == 'E' && b == 'A' ) return   -2 ;

if( a == 'E' && b == 'B' ) return    2 ;

if( a == 'E' && b == 'C' ) return   -8 ;

if( a == 'E' && b == 'D' ) return    4 ;

if( a == 'E' && b == 'E' ) return   10 ;

if( a == 'F' && b == 'A' ) return   -4 ;

if( a == 'F' && b == 'B' ) return   -6 ;

if( a == 'F' && b == 'C' ) return   -4 ;

if( a == 'F' && b == 'D' ) return   -6 ;

if( a == 'F' && b == 'E' ) return   -6 ;

if( a == 'F' && b == 'F' ) return   12 ;

if( a == 'G' && b == 'A' ) return    0 ;

if( a == 'G' && b == 'B' ) return   -2 ;

if( a == 'G' && b == 'C' ) return   -6 ;

if( a == 'G' && b == 'D' ) return   -2 ;

if( a == 'G' && b == 'E' ) return   -4 ;

if( a == 'G' && b == 'F' ) return   -6 ;

if( a == 'G' && b == 'G' ) return   12 ;

if( a == 'H' && b == 'A' ) return   -4 ;

if( a == 'H' && b == 'B' ) return    0 ;

if( a == 'H' && b == 'C' ) return   -6 ;

if( a == 'H' && b == 'D' ) return   -2 ;

if( a == 'H' && b == 'E' ) return    0 ;

if( a == 'H' && b == 'F' ) return   -2 ;

if( a == 'H' && b == 'G' ) return   -4 ;

if( a == 'H' && b == 'H' ) return   16 ;

if( a == 'I' && b == 'A' ) return   -2 ;

if( a == 'I' && b == 'B' ) return   -6 ;

if( a == 'I' && b == 'C' ) return   -2 ;

if( a == 'I' && b == 'D' ) return   -6 ;

if( a == 'I' && b == 'E' ) return   -6 ;

if( a == 'I' && b == 'F' ) return    0 ;

if( a == 'I' && b == 'G' ) return   -8 ;

if( a == 'I' && b == 'H' ) return   -6 ;

if( a == 'I' && b == 'I' ) return    8 ;

if( a == 'K' && b == 'A' ) return   -2 ;

if( a == 'K' && b == 'B' ) return    0 ;

if( a == 'K' && b == 'C' ) return   -6 ;

if( a == 'K' && b == 'D' ) return   -2 ;

if( a == 'K' && b == 'E' ) return    2 ;

if( a == 'K' && b == 'F' ) return   -6 ;

if( a == 'K' && b == 'G' ) return   -4 ;

if( a == 'K' && b == 'H' ) return   -2 ;

if( a == 'K' && b == 'I' ) return   -6 ;

if( a == 'K' && b == 'K' ) return   10 ;

if( a == 'L' && b == 'A' ) return   -2 ;

if( a == 'L' && b == 'B' ) return   -8 ;

if( a == 'L' && b == 'C' ) return   -2 ;

if( a == 'L' && b == 'D' ) return   -8 ;

if( a == 'L' && b == 'E' ) return   -6 ;

if( a == 'L' && b == 'F' ) return    0 ;

if( a == 'L' && b == 'G' ) return   -8 ;

if( a == 'L' && b == 'H' ) return   -6 ;

if( a == 'L' && b == 'I' ) return    4 ;

if( a == 'L' && b == 'K' ) return   -4 ;

if( a == 'L' && b == 'L' ) return    8 ;

if( a == 'M' && b == 'A' ) return   -2 ;

if( a == 'M' && b == 'B' ) return   -6 ;

if( a == 'M' && b == 'C' ) return   -2 ;

if( a == 'M' && b == 'D' ) return   -6 ;

if( a == 'M' && b == 'E' ) return   -4 ;

if( a == 'M' && b == 'F' ) return    0 ;

if( a == 'M' && b == 'G' ) return   -6 ;

if( a == 'M' && b == 'H' ) return   -4 ;

if( a == 'M' && b == 'I' ) return    2 ;

if( a == 'M' && b == 'K' ) return   -2 ;

if( a == 'M' && b == 'L' ) return    4 ;

if( a == 'M' && b == 'M' ) return   10 ;

if( a == 'N' && b == 'A' ) return   -4 ;

if( a == 'N' && b == 'B' ) return    6 ;

if( a == 'N' && b == 'C' ) return   -6 ;

if( a == 'N' && b == 'D' ) return    2 ;

if( a == 'N' && b == 'E' ) return    0 ;

if( a == 'N' && b == 'F' ) return   -6 ;

if( a == 'N' && b == 'G' ) return    0 ;

if( a == 'N' && b == 'H' ) return    2 ;

if( a == 'N' && b == 'I' ) return   -6 ;

if( a == 'N' && b == 'K' ) return    0 ;

if( a == 'N' && b == 'L' ) return   -6 ;

if( a == 'N' && b == 'M' ) return   -4 ;

if( a == 'N' && b == 'N' ) return   12 ;

if( a == 'P' && b == 'A' ) return   -2 ;

if( a == 'P' && b == 'B' ) return   -4 ;

if( a == 'P' && b == 'C' ) return   -6 ;

if( a == 'P' && b == 'D' ) return   -2 ;

if( a == 'P' && b == 'E' ) return   -2 ;

if( a == 'P' && b == 'F' ) return   -8 ;

if( a == 'P' && b == 'G' ) return   -4 ;

if( a == 'P' && b == 'H' ) return   -4 ;

if( a == 'P' && b == 'I' ) return   -6 ;

if( a == 'P' && b == 'K' ) return   -2 ;

if( a == 'P' && b == 'L' ) return   -6 ;

if( a == 'P' && b == 'M' ) return   -4 ;

if( a == 'P' && b == 'N' ) return   -4 ;

if( a == 'P' && b == 'P' ) return   14 ;

if( a == 'Q' && b == 'A' ) return   -2 ;

if( a == 'Q' && b == 'B' ) return    0 ;

if( a == 'Q' && b == 'C' ) return   -6 ;

if( a == 'Q' && b == 'D' ) return    0 ;

if( a == 'Q' && b == 'E' ) return    4 ;

if( a == 'Q' && b == 'F' ) return   -6 ;

if( a == 'Q' && b == 'G' ) return   -4 ;

if( a == 'Q' && b == 'H' ) return    0 ;

if( a == 'Q' && b == 'I' ) return   -6 ;

if( a == 'Q' && b == 'K' ) return    2 ;

if( a == 'Q' && b == 'L' ) return   -4 ;

if( a == 'Q' && b == 'M' ) return    0 ;

if( a == 'Q' && b == 'N' ) return    0 ;

if( a == 'Q' && b == 'P' ) return   -2 ;

if( a == 'Q' && b == 'Q' ) return   10 ;

if( a == 'R' && b == 'A' ) return   -2 ;

if( a == 'R' && b == 'B' ) return   -2 ;

if( a == 'R' && b == 'C' ) return   -6 ;

if( a == 'R' && b == 'D' ) return   -4 ;

if( a == 'R' && b == 'E' ) return    0 ;

if( a == 'R' && b == 'F' ) return   -6 ;

if( a == 'R' && b == 'G' ) return   -4 ;

if( a == 'R' && b == 'H' ) return    0 ;

if( a == 'R' && b == 'I' ) return   -6 ;

if( a == 'R' && b == 'K' ) return    4 ;

if( a == 'R' && b == 'L' ) return   -4 ;

if( a == 'R' && b == 'M' ) return   -2 ;

if( a == 'R' && b == 'N' ) return    0 ;

if( a == 'R' && b == 'P' ) return   -4 ;

if( a == 'R' && b == 'Q' ) return    2 ;

if( a == 'R' && b == 'R' ) return   10 ;

if( a == 'S' && b == 'A' ) return    2 ;

if( a == 'S' && b == 'B' ) return    0 ;

if( a == 'S' && b == 'C' ) return   -2 ;

if( a == 'S' && b == 'D' ) return    0 ;

if( a == 'S' && b == 'E' ) return    0 ;

if( a == 'S' && b == 'F' ) return   -4 ;

if( a == 'S' && b == 'G' ) return    0 ;

if( a == 'S' && b == 'H' ) return   -2 ;

if( a == 'S' && b == 'I' ) return   -4 ;

if( a == 'S' && b == 'K' ) return    0 ;

if( a == 'S' && b == 'L' ) return   -4 ;

if( a == 'S' && b == 'M' ) return   -2 ;

if( a == 'S' && b == 'N' ) return    2 ;

if( a == 'S' && b == 'P' ) return   -2 ;

if( a == 'S' && b == 'Q' ) return    0 ;

if( a == 'S' && b == 'R' ) return   -2 ;

if( a == 'S' && b == 'S' ) return    8 ;

if( a == 'T' && b == 'A' ) return    0 ;

if( a == 'T' && b == 'B' ) return   -2 ;

if( a == 'T' && b == 'C' ) return   -2 ;

if( a == 'T' && b == 'D' ) return   -2 ;

if( a == 'T' && b == 'E' ) return   -2 ;

if( a == 'T' && b == 'F' ) return   -4 ;

if( a == 'T' && b == 'G' ) return   -4 ;

if( a == 'T' && b == 'H' ) return   -4 ;

if( a == 'T' && b == 'I' ) return   -2 ;

if( a == 'T' && b == 'K' ) return   -2 ;

if( a == 'T' && b == 'L' ) return   -2 ;

if( a == 'T' && b == 'M' ) return   -2 ;

if( a == 'T' && b == 'N' ) return    0 ;

if( a == 'T' && b == 'P' ) return   -2 ;

if( a == 'T' && b == 'Q' ) return   -2 ;

if( a == 'T' && b == 'R' ) return   -2 ;

if( a == 'T' && b == 'S' ) return    2 ;

if( a == 'T' && b == 'T' ) return   10 ;

if( a == 'V' && b == 'A' ) return    0 ;

if( a == 'V' && b == 'B' ) return   -6 ;

if( a == 'V' && b == 'C' ) return   -2 ;

if( a == 'V' && b == 'D' ) return   -6 ;

if( a == 'V' && b == 'E' ) return   -4 ;

if( a == 'V' && b == 'F' ) return   -2 ;

if( a == 'V' && b == 'G' ) return   -6 ;

if( a == 'V' && b == 'H' ) return   -6 ;

if( a == 'V' && b == 'I' ) return    6 ;

if( a == 'V' && b == 'K' ) return   -4 ;

if( a == 'V' && b == 'L' ) return    2 ;

if( a == 'V' && b == 'M' ) return    2 ;

if( a == 'V' && b == 'N' ) return   -6 ;

if( a == 'V' && b == 'P' ) return   -4 ;

if( a == 'V' && b == 'Q' ) return   -4 ;

if( a == 'V' && b == 'R' ) return   -6 ;

if( a == 'V' && b == 'S' ) return   -4 ;

if( a == 'V' && b == 'T' ) return    0 ;

if( a == 'V' && b == 'V' ) return    8 ;

if( a == 'W' && b == 'A' ) return   -6 ;

if( a == 'W' && b == 'B' ) return   -8 ;

if( a == 'W' && b == 'C' ) return   -4 ;

if( a == 'W' && b == 'D' ) return   -8 ;

if( a == 'W' && b == 'E' ) return   -6 ;

if( a == 'W' && b == 'F' ) return    2 ;

if( a == 'W' && b == 'G' ) return   -4 ;

if( a == 'W' && b == 'H' ) return   -4 ;

if( a == 'W' && b == 'I' ) return   -6 ;

if( a == 'W' && b == 'K' ) return   -6 ;

if( a == 'W' && b == 'L' ) return   -4 ;

if( a == 'W' && b == 'M' ) return   -2 ;

if( a == 'W' && b == 'N' ) return   -8 ;

if( a == 'W' && b == 'P' ) return   -8 ;

if( a == 'W' && b == 'Q' ) return   -4 ;

if( a == 'W' && b == 'R' ) return   -6 ;

if( a == 'W' && b == 'S' ) return   -6 ;

if( a == 'W' && b == 'T' ) return   -4 ;

if( a == 'W' && b == 'V' ) return   -6 ;

if( a == 'W' && b == 'W' ) return   22 ;

if( a == 'X' && b == 'A' ) return    0 ;

if( a == 'X' && b == 'B' ) return   -2 ;

if( a == 'X' && b == 'C' ) return   -4 ;

if( a == 'X' && b == 'D' ) return   -2 ;

if( a == 'X' && b == 'E' ) return   -2 ;

if( a == 'X' && b == 'F' ) return   -2 ;

if( a == 'X' && b == 'G' ) return   -2 ;

if( a == 'X' && b == 'H' ) return   -2 ;

if( a == 'X' && b == 'I' ) return   -2 ;

if( a == 'X' && b == 'K' ) return   -2 ;

if( a == 'X' && b == 'L' ) return   -2 ;

if( a == 'X' && b == 'M' ) return   -2 ;

if( a == 'X' && b == 'N' ) return   -2 ;

if( a == 'X' && b == 'P' ) return   -4 ;

if( a == 'X' && b == 'Q' ) return   -2 ;

if( a == 'X' && b == 'R' ) return   -2 ;

if( a == 'X' && b == 'S' ) return    0 ;

if( a == 'X' && b == 'T' ) return    0 ;

if( a == 'X' && b == 'V' ) return   -2 ;

if( a == 'X' && b == 'W' ) return   -4 ;

if( a == 'X' && b == 'X' ) return   -2 ;

if( a == 'Y' && b == 'A' ) return   -4 ;

if( a == 'Y' && b == 'B' ) return   -6 ;

if( a == 'Y' && b == 'C' ) return   -4 ;

if( a == 'Y' && b == 'D' ) return   -6 ;

if( a == 'Y' && b == 'E' ) return   -4 ;

if( a == 'Y' && b == 'F' ) return    6 ;

if( a == 'Y' && b == 'G' ) return   -6 ;

if( a == 'Y' && b == 'H' ) return    4 ;

if( a == 'Y' && b == 'I' ) return   -2 ;

if( a == 'Y' && b == 'K' ) return   -4 ;

if( a == 'Y' && b == 'L' ) return   -2 ;

if( a == 'Y' && b == 'M' ) return   -2 ;

if( a == 'Y' && b == 'N' ) return   -4 ;

if( a == 'Y' && b == 'P' ) return   -6 ;

if( a == 'Y' && b == 'Q' ) return   -2 ;

if( a == 'Y' && b == 'R' ) return   -4 ;

if( a == 'Y' && b == 'S' ) return   -4 ;

if( a == 'Y' && b == 'T' ) return   -4 ;

if( a == 'Y' && b == 'V' ) return   -2 ;

if( a == 'Y' && b == 'W' ) return    4 ;

if( a == 'Y' && b == 'X' ) return   -2 ;

if( a == 'Y' && b == 'Y' ) return   14 ;

if( a == 'Z' && b == 'A' ) return   -2 ;

if( a == 'Z' && b == 'B' ) return    2 ;

if( a == 'Z' && b == 'C' ) return   -6 ;

if( a == 'Z' && b == 'D' ) return    2 ;

if( a == 'Z' && b == 'E' ) return    8 ;

if( a == 'Z' && b == 'F' ) return   -6 ;

if( a == 'Z' && b == 'G' ) return   -4 ;

if( a == 'Z' && b == 'H' ) return    0 ;

if( a == 'Z' && b == 'I' ) return   -6 ;

if( a == 'Z' && b == 'K' ) return    2 ;

if( a == 'Z' && b == 'L' ) return   -6 ;

if( a == 'Z' && b == 'M' ) return   -2 ;

if( a == 'Z' && b == 'N' ) return    0 ;

if( a == 'Z' && b == 'P' ) return   -2 ;

if( a == 'Z' && b == 'Q' ) return    6 ;

if( a == 'Z' && b == 'R' ) return    0 ;

if( a == 'Z' && b == 'S' ) return    0 ;

if( a == 'Z' && b == 'T' ) return   -2 ;

if( a == 'Z' && b == 'V' ) return   -4 ;

if( a == 'Z' && b == 'W' ) return   -6 ;

if( a == 'Z' && b == 'X' ) return   -2 ;

if( a == 'Z' && b == 'Y' ) return   -4 ;

if( a == 'Z' && b == 'Z' ) return    8 ;



  return -8.8;  

} /* blossom62 */



double findCM(double **x_str, int Npts, double *Tcm_str) {

  int i;

  double SumX, SumY, SumZ;

  SumX=0; SumY=0; SumZ=0;

  for(i=0; i<Npts; i++) {

    SumX+=x_str[1][i]; SumY+=x_str[2][i]; SumZ+=x_str[3][i];

  } /* for */

  Tcm_str[1] = SumX/Npts;

  Tcm_str[2] = SumY/Npts;

  Tcm_str[3] = SumZ/Npts;

  return SumZ/Npts;

}



double TRvec(double *TransVec, double **RotMatr, double *Vec) {

  double x,y,z;

  x = Vec[1]; y = Vec[2]; z = Vec[3]; 

  Vec[1] = RotMatr[1][1]*x + RotMatr[1][2]*y + RotMatr[1][3]*z;

  Vec[2] = RotMatr[2][1]*x + RotMatr[2][2]*y + RotMatr[2][3]*z;

  Vec[3] = RotMatr[3][1]*x + RotMatr[3][2]*y + RotMatr[3][3]*z;

  Vec[1] += TransVec[1];

  Vec[2] += TransVec[2];

  Vec[3] += TransVec[3];

  return Vec[3];

}



double TRmoveXYZstruc(double *TransVec, double **RotMatr, double **x_str, int Npts) {

  int i;

  double ret_val;

  double *tempVec;

  tempVec = dVector(1,3);

  for(i=0; i<Npts; i++) {

    tempVec[1] = x_str[1][i]; 

	tempVec[2] = x_str[2][i]; 

	tempVec[3] = x_str[3][i];

    ret_val = TRvec(TransVec, RotMatr, tempVec);

	x_str[1][i] = tempVec[1]; x_str[2][i] = tempVec[2]; x_str[3][i] = tempVec[3];

  }

  free_dVector(tempVec,1,3);

  return ret_val;

}



int FindMaxXYZDistIdxLast(double **x_str, int Npts) {

   int i, iMaxDist;

   double MaxDist, dist;

   MaxDist  = sqrt( SQR(x_str[1][0]) + SQR(x_str[2][0]) + SQR(x_str[3][0]));

   iMaxDist = 0;

   for(i=1; i<Npts; i++) {

     dist  = sqrt( SQR(x_str[1][i]) + SQR(x_str[2][i]) + SQR(x_str[3][i]));

	 if( dist > MaxDist) {

	   iMaxDist = i;

       MaxDist = dist;

	 } 

   }

   return iMaxDist;

}



int FindMaxXYDistIdxLast(double **x_str, int Npts) {

   int i, iMaxDist;

   double MaxDist, dist;

   MaxDist  = sqrt( SQR(x_str[1][0]) + SQR(x_str[2][0]) );

   iMaxDist = 0;

   for(i=1; i<Npts; i++) {

     dist  = sqrt( SQR(x_str[1][i]) + SQR(x_str[2][i]) );

	 if( dist > MaxDist) {

	   iMaxDist = i;

       MaxDist = dist;

	 } 

   }

   return iMaxDist;

}



double AlignZ(double x, double y, double z, double *angle_b, double *angle_c){

  *angle_b = atan( -x/z );

  *angle_c = atan( -y/(z*cos( *angle_b)-x*sin( *angle_b)) );

  return cos( *angle_b);

}



double FormRotMatrZ(double **Rtemp, double angle_a){

  Rtemp[1][1] = cos(angle_a);  Rtemp[1][2] = sin(angle_a); Rtemp[1][3] = 0.0e0; 

  Rtemp[2][1] = -sin(angle_a); Rtemp[2][2] = cos(angle_a); Rtemp[2][3] = 0.0e0; 

  Rtemp[3][1] = 0.0e0;         Rtemp[3][2] = 0.0e0;        Rtemp[3][3] = 1.0e0; 

  return Rtemp[1][1]+Rtemp[2][2]+Rtemp[3][3];

}



double FormRotMatrY(double **Rtemp, double angle_a){

  Rtemp[1][1] = cos(angle_a);  Rtemp[1][2] = 0.0e0; Rtemp[1][3] = sin(angle_a); 

  Rtemp[2][1] = 0.0e0;         Rtemp[2][2] = 1.0e0; Rtemp[2][3] = 0.0e0; 

  Rtemp[3][1] = -sin(angle_a); Rtemp[3][2] = 0.0e0; Rtemp[3][3] = cos(angle_a); 

  return Rtemp[1][1]+Rtemp[2][2]+Rtemp[3][3];

}



double FormRotMatrX(double **Rtemp, double angle_a){

  Rtemp[1][1] = 1.0e0; Rtemp[1][2] = 0.0e0;         Rtemp[1][3] = 0.0e0; 

  Rtemp[2][1] = 0.0e0; Rtemp[2][2] = cos(angle_a);  Rtemp[2][3] = sin(angle_a); 

  Rtemp[3][1] = 0.0e0; Rtemp[3][2] = -sin(angle_a); Rtemp[3][3] = cos(angle_a); 

  return Rtemp[1][1]+Rtemp[2][2]+Rtemp[3][3];

}



double MAX2(double a, double b){

  if( a>b ) return a;

  return b;

}



double MIN2(double a, double b){

  if( a<b ) return a;

  return b;

}



/*=====================================================================*/

/* Returns sum(i=0..n-1, v[3*i]) */



double vsumby3(double v[],  int n)

  {

  double sum, temp;

  int i; 



  sum = *v; n--; v += 3;

  for (i = 0; i != n; i++, v += 3) {

	  temp = *v;

	  sum += *v;

  }

  return sum;

  }



/*=====================================================================*/

/* Returns sum(i=0..n-1, v[i]^2) */

  

double vsumsqr(double v[],  int n)

  {

  double sum;

  int i; 

   

  sum = DSQR(*v); n--; v++;

  for (i = 0; i != n; i++, v++) sum += DSQR(*v);

  return sum;

  }



/*=====================================================================*/

/* Returns sum(i=0..n-1, a[3*i]*b[3*i]) */

  

double vdotby3(double a[], double b[],  int n)

  {

  double sum;

  int i; 

  

  sum = (double)(*a) * (double)(*b); n--; a += 3; b += 3;

  for (i = 0; i != n; i++, a += 3, b += 3)

    sum += (double)(*a) * (double)(*b);

  return sum;

  }





/*=====================================================================*/
